from functools import reduce
from pyspark.sql.functions import col
from pipelib.pipeline_components import DataMerger


class Joiner(DataMerger):
    """
    DataMerger that performs an equi left-outer join to merge multiple dataframes into a single one.

    Args:
        join_columns (str or list of str): The columns to be used to perform the equi-join. Must be present in each
            of the DataFrames to be merged. If there is only one column, it may be specified as a string. Otherwise,
            a list of strings must me used.

    Note:
        If columns other than the join columns share the same name in multiple DataFrames, you may face ambiguous
        reference problems. Also, if the same set of values for the join_columns occur more than once in a
        DataFrame, the result DataFrame will end up with duplicated rows.
    """
    def __init__(self, join_columns):
        super().__init__()
        self.join_columns = join_columns
        self.join_type = 'leftouter'

    def merge(self, datasets):
        """
        Perform joins over the input DataFrames to merge them into a single one.

        Args:
            datasets (pyspark.DataFrame): DataFrames to be merged

        Returns:
            The DataFrame result of the join
        """
        # print('INICIO DE UN JOINER')  # DELETE ME I'M TEMPORAL
        # for dataset in datasets:
        #     dataset.show(1000, truncate=False)

        def join(df1, df2):
            return df1.join(df2, on=self.join_columns, how=self.join_type)

        base = self._get_ids_dataframe(datasets)

        composed_dataframe = reduce(join, datasets, base)
        return composed_dataframe

    @property
    def join_columns(self):
        return self.__join_columns

    @join_columns.setter
    def join_columns(self, join_columns):
        if type(join_columns) is not list:
            self.__join_columns = [join_columns]
        else:
            self.__join_columns = join_columns

    def _get_ids_dataframe(self, dataframes):
        partial_ids_dataframes = [self._get_distinct_ids(dataframe) for dataframe in dataframes]

        def union(df1, df2):
            return df1.union(df2)

        ids_dataframe = reduce(union, partial_ids_dataframes).distinct()
        return ids_dataframe

    def _get_distinct_ids(self, dataframe):
        for column in self.join_columns:
            dataframe = dataframe.where(col(column).isNotNull())
        
        return (dataframe
                .select(*self.join_columns)
                .distinct())
