import datetime
from pyspark.sql.utils import AnalysisException
from pipelib.pipeline_components import InputProvider, OutputProvider
import pipelib.utils as utils


class CsvInputProvider(InputProvider):
    """
    Reads PySpark DataFrames from CSV files found in a providen list of paths.

    Args:
        read_paths: A list of strings containing the paths of the CSV files are going to be read.
        sql_ctx (pyspark.SQLContext): The PySpark SQLContext to be used.
    """
    def __init__(self, read_paths, sql_ctx):
        self.read_paths = read_paths
        self.sql_ctx = sql_ctx
        self.current_dataset_index = 0

    def get_dataset(self):
        """
        Reads a PySpark DataFrame from the next path in the list.

        Returns:
             The DataFrame, or None if there is no more files to read.

        Raises:
            pyspark_pipelines.utils.DataIOException: When the file can't be readen. When this occurs, the state of the
                object remains unchanged. You may want to skip the current date using the skip_read method.
        """
        if self.current_dataset_index >= len(self.read_paths):
            return None

        dataset = self._load_dataframe(self.read_paths[self.current_dataset_index])
        self.current_dataset_index += 1
        return dataset

    def skip_read(self):
        """
        Skips the current read path.

        Returns:
             None
        """
        self.current_dataset_index += 1

    def _load_dataframe(self, path):
        try:
            return utils.load_dataframe(path, self.sql_ctx)
        except AnalysisException:
            raise utils.DataIOException('%s could not load dataframe in path %s' % (self.__class__.__name__, path))


class CsvOutputProvider(OutputProvider):
    """
    Writes PySpark DataFrames as CSV files to a providen list of paths.

    Args:
        write_paths: A list of strings containing the paths where the CSV files are going to be written.
    """
    def __init__(self, write_paths, sql_ctx):
        self.write_paths = write_paths
        self.sql_ctx = sql_ctx
        self.current_dataset_index = 0

    def put_dataset(self, dataset):
        """
        Writes a Dataframe to a CSV file

        Args:
            dataset: The PySpark DataFrame to be written

        Returns:
            True if the output was successful, False if the list of paths was exhausted.
        """
        if self.current_dataset_index >= len(self.write_paths):
            return False

        write_path = self.write_paths[self.current_dataset_index]
        self._write_dataframe(dataset.coalesce(1), write_path)
        self.current_dataset_index += 1
        return True

    def skip_write(self):
        """
        Skips the current write path

        Returns:
            None
        """
        self.current_dataset_index += 1

    def _write_dataframe(self, dataframe, path):
        utils.write_dataframe(dataframe, path)


class CsvInputProviderByDate(InputProvider):
    """
    Reads PySpark DataFrames from csv files according to a date sequence.

    Seeks csv files according to the subpath structure 'year/yearmonth/yearmonthday', given a range of dates, and
    loads them as PySpark DataFrames.

    Args:
        read_path: The path from where the structured subpaths can be found.
        start_date (datetime.date): The first date to load.
        end_date (datetime.date): The last date to load.
        sql_ctx (pyspark.SQLContext): The PySpark SQLContext to be used.
    """
    def __init__(self, read_path, start_date, end_date, sql_ctx):
        super().__init__()
        self.read_path = read_path
        self.start_date = start_date
        self._current_read_date = start_date
        self.end_date = end_date
        self.sql_ctx = sql_ctx

    def get_dataset(self):
        """
        Reads the DataFrame corresponding to the next date in the date range, then advances the current date, so
        subsequent calls to this method will return different DataFrames.

        Returns:
            The DataFrame corresponding to the current date, or None if the end of the date range has been reached.

        Raises:
            pyspark_pipelines.utils.DataIOException: When the file can't be readen. When this occurs, the state of the
                object remains unchanged. You may want to skip the current date using the skip_read method.
        """
        if self._current_read_date <= self.end_date:
            read_path = self.read_path + self._get_sub_path_by_date(self._current_read_date)
            dataset = self._load_dataframe(read_path)
            self._current_read_date = self._current_read_date + datetime.timedelta(days=1)
            return dataset
        return None

    def skip_read(self):
        """
        Skips to the next date to read.

        Returns:
            None
        """
        self._current_read_date = self._current_read_date + datetime.timedelta(days=1)

    def reset(self):
        """
        Returns the object to its initial state, so that it will start again reading from the start date.

        Returns:
            None
        """
        self._current_read_date = self.start_date

    def _load_dataframe(self, path):
        try:
            return utils.load_dataframe(path, self.sql_ctx)
        except AnalysisException:
            raise utils.DataIOException('%s could not load dataframe in path %s' % (self.__class__.__name__, path))

    def _get_sub_path_by_date(self, date):
        return utils.get_date_sub_path(date)


class CsvOutputProviderByDate(OutputProvider):
    """
        Writes PySpark DataFrames to csv files according to a date sequence.

        Creates (or replace) csv files according to the subpath structure 'year/yearmonth/yearmonthday', given a range
        of dates, writing the content of PySpark DataFrames.

        Args:
            write_path: The path from where the structured subpaths should be created.
            start_date (datetime.date): The first date to write.
            end_date (datetime.date): The last date to write.
            sql_ctx (pyspark.SQLContext): The PySpark SQLContext to be used.
        """
    def __init__(self, write_path, start_date, end_date, sql_ctx):
        super().__init__()
        self.write_path = write_path
        self.start_date = start_date
        self._current_write_date = start_date
        self.end_date = end_date
        self.sql_ctx = sql_ctx

    def put_dataset(self, dataset):
        """
        Writes the DataFrame in the subpath corresponding to the current date, then advances it, so that subsequent
        calls to this method will write the dataframes in different paths, each of them corresponding to a different
        date.

        Args:
            dataset (pyspark.DataFrame): The DataFrame to be written

        Returns:
            True if the DataFrame was written, False if the end of the date range was reached and therefore the
            DataFrame was not written.

        """
        if self._current_write_date <= self.end_date:
            write_path = self.write_path + self._get_sub_path_by_date(self._current_write_date)
            self._write_dataframe(dataset, write_path)
            self._current_write_date = self._current_write_date + datetime.timedelta(days=1)
            return True
        return False

    def skip_write(self):
        """
        Skips to the next date to write.

        Returns:
            None
        """
        self._current_write_date = self._current_write_date + datetime.timedelta(days=1)

    def reset(self):
        """
        Returns the object to its initial state, so that it will start again writing from the start date.

        Returns:
            None
        """
        self._current_write_date = self.start_date

    def _write_dataframe(self, dataframe, path):
        utils.write_dataframe(dataframe, path)

    def _get_sub_path_by_date(self, date):
        return utils.get_date_sub_path(date)
