from abc import ABCMeta, abstractmethod
import logging
import pipelib.utils as utils


class Transformer(metaclass=ABCMeta):
    """
    Abtract class for a transformer.

    Transformers act as the base unit for data processing, being which transforms a dataset into another.
    """
    @abstractmethod
    def transform(self, dataset):
        """
        Trasforms a dataset according to the implementation of the transformer.

        Args:
            dataset: The data to be transformed

        Returns:
            the transformed data
        """
        pass


class Pipeline(Transformer):
    """
    Chains multiple transformation together.

    Pipelines execute multiple transformations in a sequential manner, using the output of a transformer as the input
    for the next one.

    Args:
            transformers: The list of transformers that will be used by the pipeline, in the order that the
                transformations have to be applied. Default is an empty list.
    """
    def __init__(self, transformers=None):
        if transformers is None:
            transformers = []
        super().__init__()
        self.transformers = transformers

    def add_step(self, step):
        """
        Adds a step (or Transformer) to the end of the pipeline.

        Args:
            step: A Transformer to be added at the end of the transformer list.

        Returns:
            This pipeline object, so that method chaining can be used.
        """
        self.transformers.append(step)
        return self

    def add_steps(self, step_list):
        """
        Adds a list of steps (or Transformers) to the end of the pipeline.

        Args:
            step_list: A list of Transformer to be added at the end of the transformer list.

        Returns:
            This pipeline object, so that method chaining can be used.
        """
        self.transformers += step_list
        return self

    def transform(self, dataset):
        """
        Transforms de dataset using the Transformers that have been configured to the Pipeline. This occurs in a
        sequential manner, using the output of a transformer as the input for the next one.

        Args:
            dataset: The dataset to be transformed

        Returns:
            The transformed dataset
        """
        for transformer in self.transformers:
            dataset = transformer.transform(dataset)

        return dataset


class Step(Transformer):
    """
    Performs parallel transformations on a dataset and merges the results.

    This transformer transforms a single dataset using each one of the transformers provided to it, and then merges
    the results of each into a single dataset, through the use of a DataMerger.

    Args:
        transformers: The list of Transformers to be used by the step. Default is an empty list.
        data_merger: The DataMerger to be used to merge the results of each Transformer into a single dataset.
    """
    def __init__(self, transformers=None, data_merger=None):
        if transformers is None:
            transformers = []
        super().__init__()
        self.transformers = transformers
        self.data_merger = data_merger

    def transform(self, dataset):
        """
        Transforms the input dataset using its Transformers and DataMerger.

        Args:
            dataset: The dataset to be transformed

        Returns:
             The transformed dataset
        """
        if not self.transformers or not self.data_merger:
            return dataset

        transformed_datasets = []

        for transformer in self.transformers:
            result = transformer.transform(dataset)
            transformed_datasets.append(result)

        merged_transformed_datasets = self.data_merger.merge(transformed_datasets)
        return merged_transformed_datasets

    def add_transformer(self, transformer):
        """
        Adds a transformer to the Step

        Args:
            transformer: The Transformer to be added to the Step

        Returns:
            This Step, so that method chaining can be used
        """
        self.transformers.append(transformer)
        return self


class DataMerger(metaclass=ABCMeta):
    """
    Abstract class for a DataMerger.

    DataMergers take multiple datasets, run some processing on them, and then return a single dataset as result.
    Can be thought of as a way of merging multiple sets of data into a single one.
    """
    @abstractmethod
    def merge(self, datasets):
        """
        Merges multiple datasets into a single one, according to the implementation of the DataMerger.

        Args:
            datasets (list of datasets): A list of the datasets to be merged.

        Returns:
            The merged dataset
        """
        pass


class InputProvider(metaclass=ABCMeta):
    """
    Abtract class for an InputProvider.

    An InputProvider supplies datasets that it reads, receives or generates following its own logic.
    """
    @abstractmethod
    def get_dataset(self):
        """
        Obtains a dataset and returns it. Each subsequent call to get_dataset should return a different dataset.

        Returns:
            The obtained dataset.

        Raises:
            pyspark_pipelines.utils.DataIOException: If the dataset could not be obtained. The state of the
                InputProvider must remain the same as before the call to this method.

        """
        pass

    @abstractmethod
    def skip_read(self):
        """
        Skips the next dataset to be obtained.

        Returns:
            None
        """
        pass


class OutputProvider(metaclass=ABCMeta):
    """
    Abtract class for an OutputProvider.

    An OutputProvider writes, sends, or do whatever it is implemented to do with the datasets provided to it.
    """
    @abstractmethod
    def put_dataset(self, dataset):
        """
        Receives a dataset as input and writes, sends, or outputs that dataframe in any way. This call might alter
        the state of the object, e.g. if it has a limited amount of slots to store its output.

        Args:
            dataset: The dataset to be outputted

        Returns:
            True if the output was successful, false if the OutputProvider has exhausted its potential to output
            datasets.
        """
        pass

    @abstractmethod
    def skip_write(self):
        """
        May alter the state of the OutputProvider, e.g. to skip a slot to write its output.

        Returns:
            None
        """
        pass


class Orchestrator:
    """
    Manages the reading of datasets, its processing and the output of results.

    A Orchestrator queries a InputProvider for datasets, transforms them using a Transformer, and then outputs the
    result using a OutputProvider. This sequence occurs in a loop until there is no more input to read.

    Args:
        input_provider: The provider of the input datasets.
        transformer: The Transformer to be used to transform the input datasets.
        output_provider: The OutputProvider that will output the results of the transformations.
    """
    def __init__(self, transformer, input_provider, output_provider):
        super().__init__()
        self.transformer = transformer
        self.input_provider = input_provider
        self.output_provider = output_provider

    def run(self):
        """
        Queries the InputProvider for datasets, transforms them using the Transformer, and then outputs the
        result using the OutputProvider. This sequence occurs in a loop until there is no more input to read.

        Returns:
            None

        Notes:
            This method swallows all exceptions raised during the dataset transformations, and logs them using the
            python's logging module. This is done in order to prevent a malformed dataset or other similar circumstance
            to interrupt the processing, but may prevent a bug to stop the execution. Be sure to check the logs during
            or at the end of the execution of the run method.
        """
        while True:
            try:
                input_dataset = self.input_provider.get_dataset()
            except utils.DataIOException as io_exception:
                logging.warning('The following IO exception occurred during Orchestrator.run() execution:\n\t%s\nSkipping to next dataset\n' % str(io_exception))
                self.input_provider.skip_read()
                self.output_provider.skip_write()
                continue

            if input_dataset is None:
                break

            try:
                transformed_data = self.transformer.transform(input_dataset)
            except Exception:
                logging.exception('A exception occurred during the transformation of a dataset during '
                                  + 'Orchestrator.run() execution. The data may be corrupt or you might have a bug in '
                                  + 'your transformer.\nSkipping to next day. Check the stack trace for more info.')
                self.output_provider.skip_write()
                continue

            try:
                self.output_provider.put_dataset(transformed_data)
            except utils.DataIOException as io_exception:
                logging.warning(
                    'The following IO exception occurred during Orchestrator.run() execution:\n\t%s\n'
                    + 'Skipping to next dataset\n' % io_exception)
                self.output_provider.skip_write()
                continue
