# -*- coding: utf-8 -*-
from collections import namedtuple
import logging
from pyspark.ml.feature import HashingTF
from pipelib.pipeline_components import Transformer
from pyspark.sql.functions import udf, lit, col, when, monotonically_increasing_id, collect_list, explode, first, sum
from pyspark.sql.types import IntegerType, ArrayType, StringType, DoubleType


class ColumnsSummator(Transformer):
    """
    Sums two or more numeric columns.

    Args:
        id_columns: The columns containing the id for each row.
        summations_list: A list of named tuples containing the params of each summation to be done. Default is an
            empty list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns the type for the named tuple modelling summations. Only relevant if you are going to add summations
        without the use of the helper methods.
        """
        fields = [
            'list_of_columns',
            'result_name'
        ]
        return namedtuple(cls.__name__+'Param', fields)

    def __init__(self, id_columns, summations_list=None):
        if summations_list is None:
            summations_list = []
        super().__init__()
        self.id_columns = id_columns
        self.summations_list = summations_list

    def add_sum(self, list_of_columns, result_name):
        """
        Adds a sum to be performed by the transform method.

        Args:
            list_of_columns (list of str): The name sof the columns to be summed.
            result_name: The name of the column that will hold the results of the sum.

        Returns:
            This ColumnsSummator, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(list_of_columns=list_of_columns, result_name=result_name)
        self.summations_list.append(param)
        return self

    def transform(self, dataset):
        """
        Performs all the sums over a dataset.

        Args:
            dataset: PySpark Dataframe containing the columns to be summed.

        Returns:
            A PySpark DataFrame containing the id columns and the result columns.
        """
        def summation(*args):
            accumulator = 0
            for number in args:
                if number:
                    accumulator = accumulator + int(number)
            return accumulator

        summation_udf = udf(summation, IntegerType())

        names_and_sources = {params.result_name: params.list_of_columns
                             for params in self.summations_list}

        for name, sources in names_and_sources.items():
            for source in sources:
                if not source in dataset.columns:
                    logging.warning('ColumnsSummator can not find the column "%s". It will be ignored.' % source)
                    dataset = dataset.withColumn(source, lit(0).cast(IntegerType()))
            dataset = dataset.withColumn(name, summation_udf(*sources))

        return dataset.select(*self.id_columns, *names_and_sources.keys())

    @property
    def id_columns(self):
        return self.__id_columns

    @id_columns.setter
    def id_columns(self, id_columns):
        if type(id_columns) is not list:
            self.__id_columns = [id_columns]
        else:
            self.__id_columns = id_columns


class ValueExtractor(Transformer):
    """
    Extract values from a column with strings of the kind: 'key1=value1;key2=value2'

    Args:
        id_columns: The columns containing the id for each row.
        extractions_list: A list of named tuples containing the params of each extraction to be done. Default is an
            empty list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns the type for the named tuple modelling extractions. Only relevant if you are going to add extractions
        without the use of the helper methods.
        """
        fields = [
            'source_column',
            'value_to_extract',
            'result_name'
        ]
        return namedtuple(cls.__name__ + 'Param', fields)

    def __init__(self, id_columns, extractions_list=None):
        if extractions_list is None:
            extractions_list = []
        super().__init__()
        self.id_columns = id_columns
        self.extractions_list = extractions_list

    def add_extraction(self, source_column, value_to_extract, result_name):
        """
        Adds a extraction to be performed by the transform method.

        Args:
            source_column: The column from where to extract the value.
            value_to_extract: The key corresponding to the value to extract.
            result_name: The name of the result column.

        Returns:
            This ValueExtractor, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(source_column=source_column, value_to_extract=value_to_extract, result_name=result_name)
        self.extractions_list.append(param)
        return self

    def _get_extraction_udf(self, value_to_extract):
        def extract(string: str):
            tuples = [token.split(sep='=') for token in string.split(sep=';')]
            values_dict = {tuple[0]: tuple[1] if 1 < len(tuple) else None for tuple in tuples}
            return values_dict.get(value_to_extract)
        return udf(extract, StringType())

    def transform(self, dataframe):
        """
        Performs the extractions.

        Args:
            dataframe: The dataframe containing the values to extract.

        Returns:
            A DataFrame containing the id columns and a extra column for each extraction performed.
        """
        result_columns = []

        for extraction in self.extractions_list:
            extract_udf = self._get_extraction_udf(extraction.value_to_extract)
            dataframe = dataframe.withColumn(extraction.result_name, extract_udf(extraction.source_column))
            result_columns.append(extraction.result_name)

        return dataframe.select(*self.id_columns, *result_columns)

    @property
    def id_columns(self):
        return self.__id_columns

    @id_columns.setter
    def id_columns(self, id_columns):
        if type(id_columns) is not list:
            self.__id_columns = [id_columns]
        else:
            self.__id_columns = id_columns


class ColumnsWeightedSummator(Transformer):
    """
    Sums two or more numeric columns applying a weight to each one.

    Args:
        id_columns: The columns containing the id for each row.
        summations_list: A list of named tuples containing the params of each summation to be done. Default is an
            empty list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns the type for the named tuple modelling summations. Only relevant if you are going to add summations
        without the use of the helper methods.
        """
        fields = [
            'list_of_columns',
            'list_of_weights',
            'result_name'
        ]
        return namedtuple(cls.__name__+'Param', fields)

    def __init__(self, id_columns, summations_list=None):
        if summations_list is None:
            summations_list = []
        super().__init__()
        self.id_columns = id_columns
        self.summations_list = summations_list

    def add_weighted_sum(self, list_of_columns, list_of_weights, result_name):
        """
        Adds a weighted sum to be performed by the transform method.

        Args:
            list_of_columns (list of str): The names of the columns to be summed.
            list_of_weights (list of float): The weights corresponding to each of the columns to be summed.
            result_name: The name of the column that will hold the results of the sum.

        Returns:
            This ColumnsWeightedSummator, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(list_of_columns=list_of_columns, list_of_weights=list_of_weights, result_name=result_name)
        self.summations_list.append(param)

    def _get_weighted_summation_udf(self, list_of_weights):
        def weighted_summation(*args):
            accumulator = 0
            for index, number in enumerate(args):
                accumulator = accumulator + number*list_of_weights[index]
            return accumulator

        return udf(weighted_summation, DoubleType())

    def transform(self, dataset):
        """
        Performs all the weighted sums over a dataset.

        Args:
            dataset: PySpark Dataframe containing the columns to be summed.

        Returns:
            A PySpark DataFrame containing the id columns and the result columns.
        """
        result_column_names = []

        for summation_params in self.summations_list:
            for source in summation_params.list_of_columns:
                if source not in dataset.columns:
                    logging.warning('ColumnsSummator can not find the column "%s". It will be ignored.' % source)
                    dataset = dataset.withColumn(source, lit(0).cast(IntegerType()))
            dataset = dataset.withColumn(summation_params.result_name, self._get_weighted_summation_udf(summation_params.list_of_weights)(*summation_params.list_of_columns))
            result_column_names.append(summation_params.result_name)

        return dataset.select(*self.id_columns, *result_column_names)

    @property
    def id_columns(self):
        return self.__id_columns

    @id_columns.setter
    def id_columns(self, id_columns):
        if type(id_columns) is not list:
            self.__id_columns = [id_columns]
        else:
            self.__id_columns = id_columns


class ByPass(Transformer):
    """
    Gets some columns from the input dataset, without any changes.

    Args:
        columns (List of str): The names of the columns to bypass. If none is specified, all columns will be bypassed.
    """
    def __init__(self, columns=None):
        if columns is None:
            columns = []
        super().__init__()
        self.columns = columns

    def transform(self, dataset):
        """
        Bypass some columns of a dataset.

        Args:
            dataset: The DataFrame to bypass.

        Returns:
            A DataFrame containing the bypassed columns of the input dataframe. If no columns were specified, the
                returned DataFrame will be the input DataFrame.
        """
        if not self.columns:
            return dataset
        return dataset.select(*self.columns)


class JSONFlattener(Transformer):
    """
    Gets values from a JSON DataFrame.

    Args:
        field_list: A list of named tuples with the info of each JSON field to be getted. Default is an empty list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns the type for the named tuple containing the info for each JSON field. Only relevant if you are going to
        add fields to get without the use of the helper methods.
        """
        fields = [
            'field_path',
            'result_name'
        ]
        return namedtuple(cls.__name__+'TransformationParam', fields)

    def __init__(self, field_list=None):
        if field_list is None:
            field_list = []
        super().__init__()
        self.field_list = field_list

    def add_field(self, field_path, result_name):
        """
        Adds a JSON field to be getted.

        Args:
            field_path (str): The path of the field within the JSON DataFrame.
            result_name: The name of the result column.

        Returns:
            This JSONFlattener, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(field_path=field_path, result_name=result_name)
        self.field_list.append(param)
        return self

    def transform(self, dataframe):
        """
        Gets the JSON fields from a DataFrame

        Args:
            dataframe: The JSON Dataframe from where to get the fields.

        Returns:
            A DataFrame containing just one column for each field gotten.
        """
        names_and_paths = {params.result_column_name: params.JSON_path for params in self.field_list}
        cols_to_select = [col(path).alias(path) for path in names_and_paths.values()]
        flat = dataframe.select(*cols_to_select)
        renamed = flat
        for name, path in names_and_paths.items():
            renamed = renamed.withColumnRenamed(path, name)

        return renamed


class Binarizer(Transformer):
    """
    Binarizes values, leaving them at 0 if they are empty, otherwise setting them to 1.

    Args:
        id_columns: The columns containing the id values for each row.
        column_list: A list of named tuples containing all the info about the columns to be binarized. Default is an
            empty list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns the type of the named tuples containing the info about the columns to binarize. Only relevant if you are
        not going to use the helper method.
        """
        fields = [
            'source_column',
            'result_column_name'
        ]
        return namedtuple(cls.__name__+'TransformationParam', fields)

    def __init__(self, id_columns, column_list=None):
        if column_list is None:
            column_list = []
        super().__init__()
        self.id_columns = id_columns
        self.column_list = column_list

    def add_column_to_binarize(self, column, result_name):
        """
        Adds a column to be binarized by the transform method.

        Args:
            column: The name of the column to be binarized.
            result_name: The name of the result column.

        Returns:
            This Binarizer, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(source_column=column, result_column_name=result_name)
        self.column_list.append(param)
        return self

    def transform(self, dataframe):
        """
        Binarize the input DataFrame.

        Args:
            dataframe: The DataFrame containing the columns to be binarized.

        Returns:
            A new DataFrame containing the id columns as well as the binarized columns from the input DataFrame.
        """
        def binarize(field):
            if field:
                return 1
            else:
                return 0

        binarize_udf = udf(binarize)

        names_and_sources = {params.result_column_name: params.source_column for params in self.column_list}

        cols_to_select = [binarize_udf(source).alias(name).cast(IntegerType()) for name, source in names_and_sources.items()]
        dataframe = dataframe.select(*self.id_columns, *cols_to_select)

        return dataframe

    @property
    def id_columns(self):
        return self.__id_columns

    @id_columns.setter
    def id_columns(self, id_columns):
        if type(id_columns) is not list:
            self.__id_columns = [id_columns]
        else:
            self.__id_columns = id_columns


class ColumnsSummatorByPrefix(Transformer):
    """
    Sums all columns in a DataFrame which names share a specified prefix.

    Args:
        id_columns: The columns containing the id for each row.
        summation_list: A list of named tuples containing the params of each sum to be done. Default is an empty list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns the type for the named tuple modelling sums. Only relevant if you are going to add sums without the use
        of the helper methods.
        """
        fields = [
            'prefix_of_columns_to_sum',
            'result_column_name'
        ]
        return namedtuple(cls.__name__+'TransformationParam', fields)

    def __init__(self, id_columns, summation_list=None):
        if summation_list is None:
            summation_list = []
        super().__init__()
        self.id_columns = id_columns
        self.summation_list = summation_list

    def add_summation(self, prefix, result_name):
        """
        Adds a summation to be performed by the transform method.

        Args:
            prefix: The prefix shared among the names of the columns to be summed.
            result_name: Tha name of the result column.

        Returns:
            This ColumnsSummatorByPrefix, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(prefix_of_columns_to_sum=prefix, result_column_name=result_name)
        self.summation_list.append(param)
        return self

    def transform(self, dataframe):
        """
        Performs the summations over a DataFrame.

        Args:
            dataframe: The DataFrame containing the columns to sum.

        Returns:
            A new DataFrame containing the id columns of the input DataFrame, as well as the results of each summation.
        """
        def summation(*args):
            acc = 0
            for num in args:
                acc = acc + num
            return acc

        summation_udf = udf(summation, IntegerType())

        def filter_by_prefix(strings, prefix):
            return filter(lambda s: s.startswith(prefix), strings)

        names_and_prefixes = {params.result_column_name: params.prefix_of_columns_to_sum
                              for params in self.summation_list}

        for name, prefix in names_and_prefixes.items():
            columns_to_sum = list(filter_by_prefix(dataframe.columns, prefix))
            dataframe = dataframe.withColumn(name, summation_udf(*columns_to_sum))

        return dataframe.select(*self.id_columns, *names_and_prefixes.keys())

    @property
    def id_columns(self):
        return self.__id_columns

    @id_columns.setter
    def id_columns(self, id_columns):
        if type(id_columns) is not list:
            self.__id_columns = [id_columns]
        else:
            self.__id_columns = id_columns


class RowCounterByCondition(Transformer):
    """
    Counts, for each id, the rows that fulfills a given SQL condition.

    Args:
        id_columns: The columns containing the id values for each row of the input DataFrames.
        counts_list: A list of named tuples containing the info for each count to be done. Default is an empty list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns the type for the named tuple modelling counts. Only relevant if you are going to add counts without the
        use of the helper methods.
        """
        fields = [
            'sql_condition',
            'result_column_name'
        ]
        return namedtuple(cls.__name__ + 'TransformationParam', fields)

    def __init__(self, id_columns, counts_list=None):
        if counts_list is None:
            counts_list = []
        super().__init__()
        self.id_columns = id_columns
        self.counts_list = counts_list

    def add_count(self, sql_condition, result_name):
        """
        Adds a count to be performed by the transform method.

        Args:
            sql_condition: The SQL condition that has to be fulfilled by the rows to count. Has to be expressed as a
                pyspark column condition.
            result_name: The name of the result column.

        Returns:
            This RowCounterByCondition, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(sql_condition=sql_condition, result_column_name=result_name)
        self.counts_list.append(param)
        return self

    def transform(self, dataframe):
        """
        Performs all the count over a DataFrame.

        Args:
            dataframe: The dataframe containing the rows to be counted.

        Returns:
            A new DataFrame containing the id columns as well as the result columns.
        """
        row_ids = dataframe.select(self.id_columns).distinct()
        col_names_and_conditions = {params.result_column_name: params.sql_condition
                                    for params in self.counts_list}

        for column_name, condition in col_names_and_conditions.items():
            dataframe = dataframe.withColumn(column_name, (when(condition, lit(1)).otherwise(lit(0))).cast(IntegerType()))

        cols_to_sum = [col(column) for column in col_names_and_conditions.keys()]
        dataframe = (dataframe
                     .groupBy(self.id_columns)
                     .agg(*[sum(column).alias(column) for column in col_names_and_conditions.keys()]))

        id_column_alias = 'tmp_alias_%s' % self.id_columns
        dataframe = dataframe.withColumnRenamed(self.id_columns, id_column_alias)
        dataframe = row_ids.join(dataframe, on=(dataframe[id_column_alias] == row_ids[self.id_columns]), how="left_outer")
        dataframe = dataframe.drop(id_column_alias)

        return dataframe


class ValueMapper(Transformer):
    """
    Substitutes values from a DataFrame.

    Args:
        id_columns: The columns containing the id values for each row.
        mappings_list: A list of named tuples containing all the info about the substitutions to be made. Default is an
            empty list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns the type of the named tuples containing the info about the substitutions. Only relevant if you are
        not going to use the helper method.
        """
        fields = [
            'source_column',
            'substitution_dict',
            'default_value',
            'use_default_value',
            'result_column_name'
        ]
        return namedtuple(cls.__name__ + 'TransformationParam', fields)

    def __init__(self, id_columns, mappings_list=None):
        if mappings_list is None:
            mappings_list = []
        super().__init__()
        self.id_columns = id_columns
        self.mappings_list = mappings_list

    def add_mapping(self, source_column, substitution_dict, default_value, result_name, use_default_value=True):
        """
        Add a mapping to be performed by the transform method.

        Args:
            source_column: The column which values are to be mapped.
            substitution_dict: A dictionary that specifies the substitutions to perform. Each key-value pair represents
                the original and the substitute value, respectively.
            default_value: The value which will be used to substitute a value that can't be found in the substitution
                dict.
            result_name: The name of the result column.
            use_default_value (bool):  Whether the default value should be used. If it's set to false, values that
                can't be found in the substitution dict will remain unchanged.

        Returns:
            This ValueMapper, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(source_column=source_column, substitution_dict=substitution_dict, default_value=default_value, use_default_value=use_default_value, result_column_name=result_name)
        self.mappings_list.append(param)
        return self

    def _replace_udf_builder(self, substitution_dict, default_value, use_default_value):
        def replace(item_to_replace):
            if item_to_replace in substitution_dict:
                return substitution_dict[item_to_replace]
            if use_default_value:
                return default_value
            return item_to_replace

        return udf(replace)

    def transform(self, dataset):
        """
        Performs the substitutions over an input DataFrame.

        Args:
            dataset: The DataFrame which values are to be substituted.

        Returns:
            A new DataFrame containing the id columns as well as the substituted columns.
        """
        results = [(self._replace_udf_builder(params.substitution_dict, params.default_value, params.use_default_value)(col(params.source_column))
                    .alias(params.result_column_name))
                   for params in self.mappings_list]

        return dataset.select(*self.id_columns, *results)

    @property
    def id_columns(self):
        return self.__id_columns

    @id_columns.setter
    def id_columns(self, id_columns):
        if type(id_columns) is not list:
            self.__id_columns = [col(id_columns)]
        else:
            self.__id_columns = [col(column) for column in self.id_columns]


class ValueGetter(Transformer):
    """
    Gets values from source columns in rows that fulfills a given SQL condition.

    Note:
        Gets one value for each id.

    Args:
        id_columns: The columns containing the id for each row.
        value_list: A list of named tuples containing the params of each value to be gotten. Default is an empty list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns the type for the named tuple containing the info for each value getting. Only relevant if you are going
        to add values without the use of the helper methods.
        """
        fields = [
            'value_column',
            'sql_condition',
            'result_column_name'
        ]
        return namedtuple(cls.__name__ + 'TransformationParam', fields)

    def __init__(self, id_columns, value_list=None):
        if value_list is None:
            value_list = []
        super().__init__()
        self.id_columns = id_columns
        self.value_list = value_list

    def add_value(self, value_column, sql_condition, result_name):
        """
        Adds a value to be getted by the transform method.

        Args:
            value_column: The column containing the values to be getted.
            sql_condition: The SQL condition to be fulfilled by the rows containing the values to be getted.
            result_name: The name of the result column.

        Returns:
            This ValueGetter, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(value_column=value_column, sql_condition=sql_condition, result_name=result_name)
        self.value_list.append(param)
        return self

    def transform(self, dataframe):
        """
        Gets the proper values from a DataFrame.

        Args:
            dataframe: The DataFrame containing the values to get.

        Returns:
            A new DataFrame containing the id columns of the input DataFrame, as well as the gotten values.
        """
        row_ids = dataframe.select(*self.id_columns).distinct()

        result = row_ids
        for params in self.value_list:
            values = (dataframe
                      .where(params.sql_condition)
                      .select(*self.id_columns, params.value_column)
                      .withColumnRenamed(params.value_column, params.result_column_name)
                      .groupBy(*self.id_columns)
                      .agg(first(params.result_column_name)))

            result = result.join(values, on=self.id_columns, how='leftouter')

        return result

    @property
    def id_columns(self):
        return self.__id_columns

    @id_columns.setter
    def id_columns(self, id_columns):
        if type(id_columns) is not list:
            self.__id_columns = [col(id_columns)]
        else:
            self.__id_columns = [col(column) for column in self.id_columns]


class ValueCollector(Transformer):
    """
    Collects a list of values for each id, in rows that fulfill a given SQL condition.

    Args:
        id_columns: The columns containing the id for each row.
        collecting_list: A list of named tuples containing the params of each collecting to be done. Default is an empty
            list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns the type for the named tuple with all the info about a collecting. Only relevant if you are going to add
        them without the use of the helper methods.
        """
        fields = [
            'value_column',
            'sql_condition',
            'result_column_name'
        ]
        return namedtuple(cls.__name__ + 'TransformationParam', fields)

    def __init__(self, id_columns, collecting_list=None):
        if collecting_list is None:
            collecting_list = []
        super().__init__()
        self.id_columns = id_columns
        self.collecting_list = collecting_list

    def add_collecting(self, value_column, sql_condition, result_name):
        """
        Adds a collecting to be performed by the transform method.

        Args:
            value_column: The name of the column containing the values to be collected.
            sql_condition: The sql condition to be fulfilled by the rows containing the values.
            result_name: The name of the result column.

        Returns:
            This ValueCollector, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(value_column=value_column, sql_condition=sql_condition, result_name=result_name)
        self.collecting_list.append(param)

    def transform(self, dataframe):
        """
        Collects the values from a DataFrame.

        Args:
            dataframe: The DataFrame containing the values to collect.

        Returns:
            A new DataFrame containing the id columns of the input DataFrame, with just one row for each different id
                value. The DataFrame has a new column for each collecting done, with lists of the collected values.
        """
        row_ids = dataframe.select(*self.id_columns).distinct()

        result = row_ids
        for params in self.collecting_list:
            new_column = (dataframe
                          .where(params.sql_condition or 'TRUE')
                          .groupBy(*self.id_columns)
                          .agg(*self.id_columns, collect_list(params.value_column).alias(params.result_column_name)))

            result = result.join(new_column, on=self.id_columns, how='leftouter')

        return result

    @property
    def id_columns(self):
        return self.__id_columns

    @id_columns.setter
    def id_columns(self, id_columns):
        if type(id_columns) is not list:
            self.__id_columns = [col(id_columns)]
        else:
            self.__id_columns = [col(column) for column in self.id_columns]


class OneHotEncoder(Transformer):
    """
    Performs One Hot Encoding.

    Args:
        id_columns: The columns containing the id values for each row of the input DataFrames.
        encoding_list: A list of named tuples containing the info for each encoding to be done. Default is an empty
            list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns the type for the named containing the info about a one hot encoding. Only relevant if you are going to
        add encodings without the use of the helper methods.
        """
        fields = [
            'value_column',
            'sql_condition',
            'result_columns_prefix'
        ]
        return namedtuple(cls.__name__ + 'TransformationParam', fields)

    def __init__(self, id_columns, encoding_list=None):
        if encoding_list is None:
            encoding_list = []
        super().__init__()
        self.id_columns = id_columns
        self.encoding_list = encoding_list

    def add_enconding(self, column_to_encode, sql_condition, result_prefix):
        """
        Adds a encoding to be performed by the transform method.

        Args:
            column_to_encode: The name of the column to be encoded.
            sql_condition: An sql condition used to filter rows before the enconding.
            result_prefix: The prefix in the name of the columns generated by the encoding.

        Returns:
            This OneHotEncoder, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(value_column=column_to_encode, sql_condition=sql_condition, result_columns_prefix=result_prefix)
        self.encoding_list.append(param)

    def transform(self, dataframe):
        """
        Performs all the encodings over a DataFrame.

        Args:
            dataframe: The dataframe containing the rows to be encoded.

        Returns:
            A new DataFrame containing the id columns as well as the columns resulting of the encoding.
        """
        row_ids = dataframe.select(*self.id_columns).distinct()

        result = row_ids
        for params in self.encoding_list:
            one_hot_encoding = (dataframe
                                .where(params.sql_condition or 'TRUE')
                                .groupBy(*self.id_columns)
                                .pivot(params.value_column)
                                .count())

            for column in one_hot_encoding.columns:
                if column not in self.id_columns:
                    one_hot_encoding = one_hot_encoding.withColumnRenamed(column, params.result_columns_prefix+column)

            result = (result
                      .join(one_hot_encoding, on=self.id_columns, how="left_outer")
                      .fillna(0))

        return result

    @property
    def id_columns(self):
        return self.__id_columns

    @id_columns.setter
    def id_columns(self, id_columns):
        if type(id_columns) is not list:
            self.__id_columns = [col(id_columns)]
        else:
            self.__id_columns = [col(column) for column in self.id_columns]


class TextSplitter(Transformer):
    """
    Splits text into a list of words.

    Args:
        id_columns: The columns containing the id values for each row of the input DataFrames.
        columns_to_split_list: A list of named tuples containing the info for each split to be done. Default is an empty
            list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns the type for the named tuple with the info for one text splitting. Only relevant if you are going to add
        splits without the use of the helper methods.
        """
        fields = [
            'source_column',
            'delimiter',
            'sql_condition',
            'result_column_name'
        ]
        return namedtuple(cls.__name__ + 'TransformationParam', fields)

    def __init__(self, id_columns, columns_to_split_list=None):
        if columns_to_split_list is None:
            columns_to_split_list = []
        super().__init__()
        self.id_columns = id_columns
        self.columns_to_split_list = columns_to_split_list

    def add_splitting(self, source_column, delimiter, sql_condition, result_name):
        """

        Args:
            source_column: The column containing the text to be splitted.
            delimiter: The delimiter to be used for the text split.
            sql_condition: A optional sql condition to filter the input DataFrame before the split.
            result_name: The name of the result column.

        Returns:
            This TextSplitter, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(source_column=source_column, delimiter=delimiter, sql_condition=sql_condition, result_column_name=result_name)
        self.columns_to_split_list.append(param)

    def _split_text_udf_builder(self, delimiter):
        def split_text(text: str):
            return [word.strip() for word in text.split(delimiter)]

        return udf(split_text, ArrayType(StringType()))

    def transform(self, dataframe):
        """
        Performs the splits over a DataFrame.

        Args:
            dataframe: The DataFrame containing the columns to split.

        Returns:
            A new DataFrame containing the id columns of the input DataFrame, as well as the results of each split.
        """
        row_ids = dataframe.select(*self.id_columns).distinct()
        result = row_ids

        for params in self.columns_to_split_list:
            column = (dataframe
                      .where(params.sql_condition)
                      .select(*self.id_columns, self._split_text_udf_builder(params.delimiter)(params.source_column).alias(params.result_column_name)))

            result = result.join(column, on=self.id_columns, how='leftouter')

        return result

    @property
    def id_columns(self):
        return self.__id_columns

    @id_columns.setter
    def id_columns(self, id_columns):
        if type(id_columns) is not list:
            self.__id_columns = [col(id_columns)]
        else:
            self.__id_columns = [col(column) for column in self.id_columns]


class ListExploder(Transformer):
    """
    Explodes columns containing lists.

    Generates a new row for each value in each list. The corresponding value for each row will appear in a new
    column.

    Args:
        id_columns: The columns containing the id for each row.
        columns_to_explode_list: A list of named tuples containing the params of each exploding to be done. Default is
            an empty list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns the type for the named tuple modelling explodings. Only relevant if you are going to add them without
        the use of the helper methods.
        """
        fields = [
            'column_to_explode',
            'result_column_name'

        ]
        return namedtuple(cls.__name__ + 'TransformationParam', fields)

    def __init__(self, id_columns, columns_to_explode_list=None):
        if columns_to_explode_list is None:
            columns_to_explode_list = []
        super().__init__()
        self.id_columns = id_columns
        self.columns_to_explode_list = columns_to_explode_list

    def add_exploding(self, column_to_explode, result_name):
        """
        Adds a exploding to be performed by the transform method.

        Args:
            column_to_explode: The row containing the lists to be exploded.
            result_name: The name of the column that will contain the exploded values.

        Returns:
            This ListExploder, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(column_to_explode=column_to_explode, result_name=result_name)
        self.columns_to_explode_list.append(param)

    def transform(self, dataframe):
        """
        Performs the explodings over a DataFrame.

        Note:
            After the explosion of a list, it's likely that there is repeated id values across multiple rows.

        Args:
            dataframe: The DataFrame containing the lists to explode.

        Returns:
            A new DataFrame containing the id columns of the input DataFrame, as well as the results of each exploding.
        """
        row_ids = dataframe.select(*self.id_columns).distinct()
        result = row_ids

        for params in self.columns_to_explode_list:
            new_column = dataframe.select(*self.id_columns, explode(col(params.column_to_explode)).alias(params.result_column_name))
            result = result.join(new_column, on=self.id_columns, how='leftouter')

        return result

    @property
    def id_columns(self):
        return self.__id_columns

    @id_columns.setter
    def id_columns(self, id_columns):
        if type(id_columns) is not list:
            self.__id_columns = [col(id_columns)]
        else:
            self.__id_columns = [col(column) for column in self.id_columns]


class FeatureHasher(Transformer):
    """
    Performs feature hashing (also known as the hashing trick).

    Args:
        id_columns: The columns containing the id for each row.
        columns_to_hash_list: A list of named tuples containing the params of each hashing to be done. Default is an
            empty list.

    .. _More info:
            https://en.wikipedia.org/wiki/Feature_hashing
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns the type for the named tuple with the info about one hashing. Only relevant if you are going to add them
        without the use of the helper methods.
        """
        fields = [
            'column_to_hash',
            'number_of_features',
            'result_columns_prefix'
        ]
        return namedtuple(cls.__name__ + 'TransformationParam', fields)

    def __init__(self, id_columns, columns_to_hash_list=None):
        if columns_to_hash_list is None:
            columns_to_hash_list = []
        super().__init__()
        self.id_columns = id_columns
        self.columns_to_hash_list = columns_to_hash_list

    def add_hashing(self, column_to_hash, number_of_features, result_columns_prefix):
        """
        Adds a hashing to be performed by the transform method.

        Args:
            column_to_hash: The column to hash.
            number_of_features: The number of columns to generate. Should be a power of two.
            result_columns_prefix: A common prefix for the names of the columns to be generated by the feature hashing.

        Returns:
            This FeatureHasher, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(column_to_hash=column_to_hash, number_of_features=number_of_features, result_columns_prefix=result_columns_prefix)
        self.columns_to_hash_list.append(param)

    def transform(self, dataframe):
        """
        Performs the feature hashing over a DataFrame.

        Args:
            dataframe: The DataFrame containing the feature columns to hash.

        Returns:
            A new DataFrame containing the id columns of the input DataFrame, as well as the results of each feature
                hashing.
        """
        input_schema = dataframe.columns

        feature_index = 0
        for params in self.columns_to_hash_list:
            hasher = HashingTF(numFeatures=params.number_of_features, inputCol=params.column_to_hash, outputCol=params.result_columns_prefix+'_feature_'+feature_index)
            dataframe = hasher.transform(dataframe).select(*self.id_columns, params.result_columns_prefix+'_feature_'+feature_index)
            feature_index = feature_index + 1

        for column in input_schema:
            if column not in self.id_columns:
                dataframe = dataframe.drop(column)

        return dataframe

    @property
    def id_columns(self):
        return self.__id_columns

    @id_columns.setter
    def id_columns(self, id_columns):
        if type(id_columns) is not list:
            self.__id_columns = [col(id_columns)]
        else:
            self.__id_columns = [col(column) for column in self.id_columns]


class ValueSummatorById(Transformer):
    """
    Sums all the values in a column belonging to each id.

    Generates a new column with one row for each different id value, containing the sum of the falues found in a
    numeric column for that value.

    Args:
        id_columns: The columns containing the id values for each row.
        summation_list: A list of named tuples containing all the info about the sums to be done. Default is an empty
            list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns the type of the named tuples containing the info about a sum. Only relevant if you are not going to use
        the helper method.
        """
        fields = [
            'value_column',
            'sql_condition',
            'result_columnn_name'
        ]
        return namedtuple(cls.__name__ + 'TransformationParam', fields)

    def __init__(self, id_columns, summation_list=None):
        if summation_list is None:
            summation_list = []
        super().__init__()
        self.id_columns = id_columns
        self.summation_list = summation_list

    def add_sum(self, value_column, sql_condition, result_name):
        """
        Adds a sum to be performed by the transform method.

        Args:
            value_column: The column containing the numeric values to be summed.
            sql_condition: A sql condition used to filter rows before the sum.
            result_name: The name of the column containing the sum of values for each id.

        Returns:
            This ValueSummatorById, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(value_column=value_column, sql_condition=sql_condition, result_name=result_name)
        self.summation_list.append(param)

    def transform(self, dataframe):
        """
        Performs the sums over the input DataFrame.

        Args:
            dataframe: The DataFrame containing the columns to be summed.

        Returns:
            A new DataFrame containing the id columns as well as the result columns.
        """
        row_ids = dataframe.select(*self.id_columns).distinct()
        result = row_ids
        for params in self.summation_list:
            new_column = (dataframe
                          .where(params.sql_condition or 'TRUE')
                          .groupBy(*self.id_columns)
                          .agg(sum(params.value_column).alias(params.result_columnn_name)))

            result = result.join(new_column, on=self.id_columns, how='leftouter')

        return result.fillna(0)

    @property
    def id_columns(self):
        return self.__id_columns

    @id_columns.setter
    def id_columns(self, id_columns):
        if type(id_columns) is not list:
            self.__id_columns = [col(id_columns)]
        else:
            self.__id_columns = [col(column) for column in self.id_columns]


class StringNormalizer(Transformer):
    """
    Change to lowercase and strips a string from whitespaces.

    Args:
        id_columns: The columns containing the id values for each row.
        columns_to_normalize: The names of the columns to be normalized.
    """
    def __init__(self, id_columns, columns_to_normalize=None):
        if columns_to_normalize is None:
            columns_to_normalize = []
        super().__init__()
        self.id_columns = id_columns
        self.columns_to_normalize = columns_to_normalize

    def add_normalization(self, column):
        """
        Adds a column to be string normalized by the transform method.

        Args:
            column: The name of the column to be normalized

        Returns:
            This StringNormalizer, so that method chaining can be used.
        """
        self.columns_to_normalize.append(column)
        return self

    def _get_text_normalizer_udf(self):
        def normalize_text(text):
            return text.strip().lower()

        return udf(normalize_text, StringType())

    def transform(self, dataframe):
        """
        Normalizes the given text fields of a DataFrame

        Args:
            dataframe: The Dataframe from where to normalize the text fields.

        Returns:
            A DataFrame containing the same columns of the input DataFrame, but with the given columns with its text
                normalized.
        """
        normalize_text_udf = self._get_text_normalizer_udf()

        for column_to_normalize in self.columns_to_normalize:
            tmp_alias = 'tmp_alias_%s' % column_to_normalize
            dataframe = (dataframe
                         .withColumn(tmp_alias, normalize_text_udf(column_to_normalize))
                         .drop(column_to_normalize)
                         .withColumnRenamed(tmp_alias, column_to_normalize))

        return dataframe


class Identifier(Transformer):
    """
    Adds a unique id column to a DataFrame.

    Args:
        id_columns_name: The name of the new id column
    """
    def __init__(self, id_columns_name):
        super().__init__()
        self._uuid_columns_name = id_columns_name

    def transform(self, dataframe):
        """
        Generates a new unique id column for an input DataFrame

        Args:
            dataframe: The input DataFrame

        Returns:
            A DataFrame like the input DataFrame, with a the new unique id column.
        """
        dataframe = dataframe.withColumn(self._uuid_columns_name, monotonically_increasing_id())
        return dataframe


class NaFiller(Transformer):
    """
    Substitutes null values with a given value.

    fill_params: {
        'nombre de columna': valor por el que substituir los nulos
    }

    Args:
        fill_params: A dictionary containing a key-value pair for each column which null values has to be filled.
            Each key is a column name, and its value the corresponding fill value.
    """
    def __init__(self, fill_params):
        super().__init__()
        self._fill_params = fill_params

    def transform(self, dataframe):
        """
        Fills the nulls of a DataFrame

        Args:
            dataframe: The DataFrame to fill.

        Returns:
            The filled DataFrame.
        """
        return dataframe.fillna(self._fill_params)


class ColumnDropper(Transformer):
    """
    Drops columns from a DataFrame

    Args:
        cols_to_drop: A list of the names of the columns to be dropped.
    """
    def __init__(self, cols_to_drop):
        super().__init__()
        self._cols_to_drop = cols_to_drop

    def transform(self, dataframe):
        """
        Drops the given columns from a input DataFrame.

        Args:
            dataframe: The input DataFrame.

        Returns:
            The DataFrame with the columns dropped.
        """
        for col in self._cols_to_drop:
            dataframe = dataframe.drop(col)

        return dataframe


class Function(Transformer):
    """
    Applies a function to all the values in a column
    """

    @classmethod
    def get_param_type(cls):
        """
        Returns the type of the named tuples containing the info about a function to apply. Only relevant if you are not
        going to use the helper method.
        """
        fields = [
            'source_column',
            'function',
            'result_columnn_name'
        ]
        return namedtuple(cls.__name__ + 'TransformationParam', fields)

    def __init__(self, id_columns, functions_list=None):
        raise NotImplementedError('The Function transformer has not been implemented yet')
        if functions_list is None:
            functions_list = []
        super().__init__()
        self.id_columns = id_columns
        self.functions_list = functions_list

    def add_function(self, source_column, function, result_name):
        Param = self.get_param_type()
        param = Param(source_column=source_column, function=function, result_column_name=result_name)
        self.functions_list.append(param)
        return self

    def transform(self, dataframe):
        pass