import datetime


def get_date_sub_path(date: datetime.date):
    """
    Generates a subpath from a date, according to the structure: year/yearmonth/yearmonthday

    Args:
        date (datetime.date): The date used for the subpath generation.

    Returns:
        The subpath corresponding to the input date.
    """
    y = date.year
    m = date.month
    d = date.day
    return '%04d/%04d%02d/%04d%02d%02d/' % (y, y, m, y, m, d)


def load_dataframe(read_path, sqlCtx, delimiter=";", mode="DROPMALFORMED", schema=None):
    """
    Loads a PySpark DataFrame from a csv file.

    Args:
        read_path: The path where the csv data can be found.
        sqlCtx: The PySpark SQLContext to use for the read.
        delimiter: The delimiter of the csv file. Default is ';'.
        mode: The mode of the read. Default is 'DROPMALFORMED'. More modes can be found on PySpark documentation.
        schema (pyspark.sql.types.StructType): The schema of the data. If none is provided, it will be inferred.

    Returns:
        The loaded DataFrame.
    """
    if schema:
        return sqlCtx.read.load(read_path,
                                format='com.databricks.spark.csv',
                                header=True,
                                schema=schema,
                                delimiter=delimiter,
                                encoding='UTF-8',
                                mode=mode)
    else:
        return sqlCtx.read.load(read_path,
                                format='com.databricks.spark.csv',
                                header='true',
                                inferSchema=True,
                                delimiter=delimiter,
                                encoding='UTF-8',
                                mode=mode)


def write_dataframe(dataframe, path):
    """
    Writes a PySpark dataframe to a csv file.

    Args:
        dataframe: The dataframe to be written.
        path: The path were the dataframe have to be written.

    Returns:
        None
    """
    dataframe.write.csv(path, header=True, sep=";", mode='overwrite')


class DataIOException(Exception):
    """
    A exception raised by an Input or Output Provider when an error occurs.
    """
    pass
