from pyspark.sql.types import StringType
from pyspark.sql.functions import col
from pipelib.pyspark_step import PysparkStep
from tealeaf.transformers import *


class TealeafStep(PysparkStep):
    """
    Performs multiple transformations over a single PySpark Dataframe, and then merges the results into a single one.

    This Step provides a easy to use interface to configure transformations over a single PySpark Dataframe, runs
    them, and then merges the results using a DataMerger.

    Args:
        transformers: The list of Transformers to be used by the step. Default is an empty list. If you are going to use
            the easy to use interface to configure the transformations, you may want to leave it as its default value.
        data_merger: The DataMerger to be used to merge the results of each transformation into a single result
            DataFrame.
        session_id_column: The name of the column that contains the session id values in the input DataFrame.
        event_id_column: The name of the column containing the event id values.
        row_type_column: The name of the column containing the row type values.
        hit_key_column: The name of the column containing the hit key values.
        fact_value_column: The name of the column containing the fact values of events.
        timestamp_column: The name of the column containing the timestamp of the timestamped hits.
        report_group_type_column: The name of the column containing the type of the report groups.
    """
    def __init__(self, data_merger, transformers=None, session_id_column='SESSION_KEY', event_id_column='EVENT_ID',
                 row_type_column='ROW_TYPE', hit_key_column='HIT_KEY', fact_value_column='FACT_VALUE',
                 timestamp_column='EVENT_TIMESTAMP', report_group_type_column='GROUP_TYPE',
                 report_group_id_column='RG_ID', dim_0_column='FACT_DIM_0', dim_1_column='FACT_DIM_1',
                 dim_2_column='FACT_DIM_2', dim_3_column='FACT_DIM_3'):
        if transformers is None:
            transformers = []
        super().__init__(transformers=transformers, data_merger=data_merger, id_column=session_id_column)
        self.event_id_column = event_id_column
        self.row_type_column = row_type_column
        self.hit_key_column = hit_key_column
        self.fact_value_column = fact_value_column
        self.timestamp_column = timestamp_column
        self.report_group_type_column = report_group_type_column
        self.report_group_id_column = report_group_id_column
        self.dim_columns = [dim_0_column, dim_1_column, dim_2_column, dim_3_column]

    def slice_sessions(self, slicer_event_id, subsession_id_column_name):
        """
        Breaks sessions into smaller pieces.

        Labels each row with a 'subsession' id, splitting each session into multiple subsessions. This is done using a
        slicing event as point of subsession change. e.g.: if the slicing event occurs at time 5 and time 10, all
        other events in that session occurring between times 0 and 5 will belong to subsession 1; al events between
        times 6 and 10 will belong to subsession 2; and all events that occurs past time 10 will belong to subsession 3.

        Args:
            slicer_event_id: The id of the slicer event.
            subsession_id_column_name: The name of the column containing the subsession id.

        Returns:
            This step, so that method chaining can be used.
        """
        slicer = SessionSlicer(session_id_column=self.id_column,
                               event_id_column=self.event_id_column,
                               hit_key_column=self.hit_key_column,
                               report_group_type_column=self.report_group_type_column,
                               timestamp_column=self.timestamp_column,
                               slicer_event_id=slicer_event_id,
                               result_name=subsession_id_column_name)
        self.transformers.append(slicer)
        return self

    def count_event(self, event_id, result_name, result_type=StringType()):
        """
        Counts how many events of a given type occurred to each session.

        Args:
            event_id (str): The id of the event to be counted.
            result_name: The name of the column containing the count.

        Returns:
            This step, so that method chaining can be used.
        """
        if EventCounter not in self._get_transformer_types():
            event_counter = EventCounter(session_id_column=self.id_column,
                                         event_id_column=self.event_id_column,
                                         row_type_column=self.row_type_column)
            self.transformers.append(event_counter)
            self._transformer_dict[EventCounter] = event_counter
        self._transformer_dict[EventCounter].add_event(event_id=event_id, result_name=result_name)
        return self

    def check_event(self, event_id, result_name, result_type=StringType()):
        """
        Checks if a given event occurs, for each session.

        Args:
            event_id (str): The id of the event to be checked.
            result_name: The name of the result column.

        Returns:
            This step, so that method chaining can be used.
        """
        if EventChecker not in self._get_transformer_types():
            event_checker = EventChecker(session_id_column=self.id_column,
                                         event_id_column=self.event_id_column,
                                         row_type_column=self.row_type_column)
            self.transformers.append(event_checker)
            self._transformer_dict[EventChecker] = event_checker
        self._transformer_dict[EventChecker].add_event(event_id=event_id, result_name=result_name)
        return self

    def get_last_fact_value(self, event_id, result_name, result_type=StringType()):
        """
        Gets the fact value of the last occurrence of a given event for each session.

        Note:
            The event must be timestamped.

        Args:
            event_id (str): The id of the timestamped event to be used.
            result_name: The name of the result column.

        Returns:
            This step, so that method chaining can be used.
        """
        if LastFactValueFinder not in self._get_transformer_types():
            last_value_finder = LastFactValueFinder(session_id_column=self.id_column,
                                                    event_id_column=self.event_id_column,
                                                    hit_key_column=self.hit_key_column,
                                                    fact_value_column=self.fact_value_column,
                                                    timestamp_column=self.timestamp_column)
            self.transformers.append(last_value_finder)
            self._transformer_dict[LastFactValueFinder] = last_value_finder
        self._transformer_dict[LastFactValueFinder].add_event(event_id=event_id, result_name=result_name)
        return self

    def count_fact_value(self, event_id, value_to_count, result_name, result_type=StringType()):
        """
        For each id, counts the number of appearances of a given fact value in a given event.

        Args:
            event_id: The id of the event.
            value_to_count: The fact value to be counted.
            result_name: The name of the result column.

        Returns:
            This step, so that method chaining can be used.
        """
        if FactValueCounter not in self._get_transformer_types():
            value_counter = FactValueCounter(session_id_column=self.id_column,
                                             event_id_column=self.event_id_column,
                                             row_type_column=self.row_type_column)
            self.transformers.append(value_counter)
            self._transformer_dict[FactValueCounter] = value_counter
        self._transformer_dict[FactValueCounter].add_value_to_count(event_id=event_id, fact_value=value_to_count,
                                                                    result_name=result_name)
        return self

    def check_fact_value(self, event_id, value_to_check, result_name, result_type=StringType()):
        """
        Checks if a fact value appears in a given event type for each session.

        Args:
            event_id: The id of the event which fact values are to be checked.
            value_to_check: The fact value to check.
            result_name: Te name of the result column.

        Returns:
            This step, so that method chaining can be used.
        """
        if FactValueChecker not in self._get_transformer_types():
            value_checker = FactValueChecker(session_id_column=self.id_column,
                                             event_id_column=self.event_id_column,
                                             row_type_column=self.row_type_column)
            self.transformers.append(value_checker)
            self._transformer_dict[FactValueChecker] = value_checker
        self._transformer_dict[FactValueChecker].add_value_to_check(event_id=event_id, fact_value=value_to_check,
                                                                    result_name=result_name)
        return self

    def get_fact_value(self, event_id, result_name, result_type=StringType()):
        """
        Gets the fact value of a given event, for each session.

        If that event occurs more that once for a session, any of the fact values will be taken, and just one of them.

        Args:
            event_id: The id of the event.
            result_name: The name of the result column.

        Returns:
            This step, so that method chaining can be used.
        """
        if FactValueFinder not in self._get_transformer_types():
            value_finder = FactValueFinder(session_id_column=self.id_column,
                                           event_id_column=self.event_id_column)
            self.transformers.append(value_finder)
            self._transformer_dict[FactValueFinder] = value_finder
        self._transformer_dict[FactValueFinder].add_event(event_id=event_id, result_name=result_name)
        return self

    def get_dimension_value(self, event_id, report_group_id, dimension_index, result_name, result_type=StringType()):
        """
        Gets a dimension value from a report group.

        Args:
            event_id: The id of the event holding the required report group.
            report_group_id: The id of the report group.
            dimension_index: The index, from 0 to 3, of the dimension to get.
            result_name: The name of the result column.

        Returns:
            This step, so that method chaining can be used.
        """
        dimension_column = self.dim_columns[dimension_index]
        condition = '%s = "%s" AND %s = "%s"' % (self.event_id_column, event_id,
                                                 self.report_group_id_column, report_group_id)

        self.get_value(source_column=dimension_column, result_name=result_name, sql_condition=condition,
                       result_type=result_type)

    def count_dimension_value(self, event_id, report_group_id, dimension_index, value_to_count, result_name,
                              result_type=StringType()):
        """
        Counts the appearances of a value in a dimension of a report group.

        Args:
            event_id: The id of the event holding the report group.
            report_group_id: The id of the report group.
            dimension_index: The index, from 0 to 3, of the dimension where the value will be counted.
            value_to_count: The value to be counted.
            result_name: The name of the result column.

        Returns:
            This step, so that method chaining can be used.
        """
        dimension_column = self.dim_columns[dimension_index]

        # sql_condition = '%s = "%s" AND %s = "%s" AND %s = "%s"' % (self.event_id_column, event_id,
        #                                                            self.report_group_id_column, report_group_id,
        #                                                            dimension_column, value_to_count)
        sql_condition = (col(self.event_id_column) == event_id) & (col(self.report_group_id_column) == report_group_id) & (col(dimension_column) == value_to_count)
        self.count_rows_by_condition(sql_condition=sql_condition, result_name=result_name, result_type=result_type)
        return self
