from collections import namedtuple
from functools import reduce
from pyspark.sql.functions import col, lit, udf, first, concat, rank, lag, max
from pyspark.sql.types import StringType, IntegerType
from pipelib.pipeline_components import Transformer
from pyspark.sql.window import Window


class EventCounter(Transformer):
    """
    Counts event occurrences for each session.

    Args:
        session_id_column: The column containing the session ids.
        event_id_column: The column containing the event ids.
        row_type_column: The column containing the row type.
        param_list: A list of named tuples containing the necessary params for each event count. Default value is an
            empty list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns a named tuple type with all the params necessary to perform an event count. Ignore this if you are going
        to use the helper methods.
        """
        fields = [
            'event_id',
            'result_column_name'
        ]
        return namedtuple(cls.__name__+'Param', fields)

    def __init__(self, session_id_column, event_id_column, row_type_column, param_list=None):
        if param_list is None:
            param_list = []
        super().__init__()
        self.session_id_column = session_id_column
        self.event_id_column = event_id_column
        self.row_type_column = row_type_column
        self.param_list = param_list

    def add_event(self, event_id, result_name):
        """
        Add a event to be counted

        Args:
            event_id (str): The id of the event to be counted.
            result_name: The name of the result column containing the count.

        Returns:
            This EventCounter, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(event_id=event_id, result_column_name=result_name)
        self.param_list.append(param)

    def transform(self, dataframe):
        """
        Performs the event count.

        Args:
            dataframe: A DataFrame containing the events to be counted.

        Returns:
            A new DataFrame containing the session id, as well as the count of each of the events for each session.
        """
        event_id_column = self.event_id_column
        session_id_column = self.session_id_column
        sesiones = dataframe.select(*session_id_column).distinct()
        eventos = dataframe.where(col(self.row_type_column) == "EVENT")

        # Genera la condición para filtrar las filas de los eventos que nos interesan
        hay_que_contar_este_evento = reduce(
            lambda condicion1, condicion2: condicion1+" OR "+condicion2,
            [str(event_id_column)+' = "'+str(params.event_id)+'"' for params in self.param_list]
        )

        # Hace la cuenta de eventos para cada sesión
        event_counts_by_session = (eventos
                                   .where(hay_que_contar_este_evento)
                                   .groupBy(*session_id_column)
                                   .pivot(event_id_column)
                                   .count())

        result = (sesiones
                  .join(event_counts_by_session, on=session_id_column, how="left_outer")
                  )

        # Da a las columnas los nombres de los eventos
        for params in self.param_list:
            if str(params.event_id) in result.columns:
                result = (result
                          .withColumnRenamed(str(params.event_id), params.result_column_name)
                          .fillna({params.result_column_name: 0}))
            else:
                result = result.withColumn(params.result_column_name, lit(0))

        return result

    @property
    def session_id_column(self):
        return self.__session_id_column

    @session_id_column.setter
    def session_id_column(self, session_id_column):
        if type(session_id_column) is not list:
            self.__session_id_column = [session_id_column]
        else:
            self.__session_id_column = session_id_column


class EventChecker(Transformer):
    """
    Checks if a certain event occurs for each session.

    Args:
        session_id_column: The column containing the session ids.
        event_id_column: The column containing the event ids.
        row_type_column: The column containing the row type.
        param_list: A list of named tuples containing the necessary params for each event check. Default value is an
            empty list.
    """
    @classmethod
    def get_param_type(cls):
        """
         Returns a named tuple type with all the params necessary to perform an event check. Ignore this if you are
         going to use the helper methods.
         """
        fields = [
            'event_id',
            'result_column_name'
        ]
        return namedtuple(cls.__name__ + 'Param', fields)

    def __init__(self, session_id_column, event_id_column, row_type_column, param_list=None):
        if param_list is None:
            param_list = []
        super().__init__()
        self.session_id_column = session_id_column
        self.event_id_column = event_id_column
        self.row_type_column = row_type_column
        self.param_list = param_list

    def add_event(self, event_id, result_name):
        """
        Add a event to be checked.

        Args:
            event_id (str): The id of the event to be checked.
            result_name: The name of the result column containing the check.

        Returns:
            This EventChecker, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(event_id=event_id, result_column_name=result_name)
        self.param_list.append(param)

    def transform(self, dataframe):
        """
        Performs the event check.

        Args:
            dataframe: A DataFrame containing the events to be checked.

        Returns:
             A new DataFrame containing the session id, as well as the check of each of the events for each session.
        """
        event_id_column = self.event_id_column
        session_id_column = self.session_id_column
        sessions_df = dataframe.select(*session_id_column).distinct()
        events_df = dataframe.where(col(self.row_type_column) == "EVENT")

        events_to_check = {str(params.event_id): params.result_column_name
                           for params in self.param_list}

        # Genera la condición para filtrar las filas de los eventos que nos interesan
        hay_que_checkear_este_evento = reduce(
            lambda condicion1, condicion2: condicion1 + " OR " + condicion2,
            [event_id_column + ' = "' + event_id+'"' for event_id in events_to_check]
        )
        events_df = events_df.where(hay_que_checkear_este_evento)

        booleanizer = udf(lambda x: 0 if x is None else 1)

        events_df = (events_df
                     .groupBy(*session_id_column)
                     .pivot(event_id_column)
                     .agg(col(event_id_column).isNotNull()))

        events_df = sessions_df.join(events_df, on=session_id_column, how="left_outer")

        for event_id, column_name in events_to_check.items():
            if event_id in events_df.columns:
                events_df = (events_df
                             .withColumnRenamed(event_id, event_id + "_tmp")
                             .withColumn(column_name, booleanizer(event_id + "_tmp"))
                             .drop(event_id + "_tmp"))
            else:
                events_df = events_df.withColumn(column_name, lit(0))
        return events_df

    @property
    def session_id_column(self):
        return self.__session_id_column

    @session_id_column.setter
    def session_id_column(self, session_id_column):
        if type(session_id_column) is not list:
            self.__session_id_column = [session_id_column]
        else:
            self.__session_id_column = session_id_column
    

class SessionSlicer(Transformer):
    """
    Breaks sessions into smaller pieces.

    Labels each row with a 'subsession' id, splitting each session into multiple subsessions. This is done using a
    slicing event as point of subsession change. e.g.: if the slicing event occurs at time 5 and time 10, all
    other events in that session occurring between times 0 and 5 will belong to subsession 1; al events between
    times 6 and 10 will belong to subsession 2; and all events that occurs past time 10 will belong to subsession 3.

    Notes:
        The timestamps must be of type pyspark.sql.types.IntegerType

    Args:
        session_id_column: The column containing the session ids.
        event_id_column: The column containing the event ids.
        hit_key_column: The column containing the hit keys.
        report_group_type_column: The column containing the report group type.
        timestamp_column: The column containing the timestamp of type IntegerType.
        slicer_event_id: The id of the event used to do the slicing. Default value is None.
        result_name: The name of the subsession column. Default value is None.

    Note:
        If you don't specify the values of slicer_event_id or result_name, you will have to do so before calling the
        transform method, either manually or using the set_slicing method.
    """

    def __init__(self, session_id_column, event_id_column, hit_key_column, report_group_type_column,
                 timestamp_column, slicer_event_id=None, result_name=None):
        super().__init__()
        self.session_id_column = session_id_column
        self.event_id_column = event_id_column
        self.hit_key_column = hit_key_column
        self.timestamp_column = timestamp_column
        self.report_group_type_column = report_group_type_column
        self.slicer_event_id = slicer_event_id
        self.result_name = result_name

    def set_slicing(self, slicer_event_id, result_name):
        """
        Sets the slicer event id and result column name.
        """
        self.slicer_event_id = slicer_event_id
        self.result_name = result_name

    def transform(self, dataframe):
        """
        Slices the sesions of a DataFrame

        Args:
            dataframe: The dataframe which sessions are to be sliced.

        Returns:
            The input DataFrame, with a new column specifying the subsession id for each row. For non timestamped
                rows, the subsesion id will be null.
        """
        session_window = (Window
                          .partitionBy(self.session_id_column)
                          .orderBy(col(self.timestamp_column).asc()))

        start_timestamp_column = 'tmp_start_%s' % self.timestamp_column

        explicit_session_slices = (dataframe
                                   .where(col(self.event_id_column) == self.slicer_event_id)
                                   .where(col(self.report_group_type_column) == 'TIMESTAMP')
                                   .select(col(self.session_id_column), col(self.timestamp_column).cast(IntegerType()))
                                   .withColumn(self.result_name, rank().over(session_window))
                                   .withColumn(start_timestamp_column,
                                               lag(self.timestamp_column, count=1, default=-1).over(session_window).cast(IntegerType())))

        upper_unbounded_slices = (explicit_session_slices
                                  .groupBy(self.session_id_column)
                                  .agg((max(self.result_name) + lit(1)).alias(self.result_name),
                                       lit(2147483647).alias(self.timestamp_column),
                                       max(self.timestamp_column).alias(start_timestamp_column))
                                  .select(*explicit_session_slices.columns))

        session_column_alias = 'tmp_alias_%s' % self.session_id_column  # Necessary according to Spark Warn
        session_slices = (explicit_session_slices
                          .union(upper_unbounded_slices)
                          .withColumnRenamed(self.session_id_column, session_column_alias)
                          .distinct())

        timestamped_hits = (dataframe
                            .where(col(self.report_group_type_column) == 'TIMESTAMP')
                            .select(col(self.session_id_column), col(self.hit_key_column),
                                    col(self.timestamp_column).cast(IntegerType()))
                            .groupBy(col(self.session_id_column), col(self.hit_key_column))
                            .agg(first(col(self.timestamp_column)).alias(self.timestamp_column)))

        join_condition = ((timestamped_hits[self.session_id_column] == session_slices[session_column_alias])
                          & (timestamped_hits[self.timestamp_column] <= session_slices[self.timestamp_column])
                          & (timestamped_hits[self.timestamp_column] > session_slices[start_timestamp_column]))

        hits_with_subsession_id = (timestamped_hits
                                   .join(session_slices, on=join_condition, how='leftouter')
                                   .select(timestamped_hits[self.session_id_column],
                                           timestamped_hits[self.hit_key_column],
                                           session_slices[self.result_name]))

        # Not timestamped hits with all the subsession ids of their session
        not_timestamped_hits = (dataframe
                                .select(col(self.hit_key_column), col(self.session_id_column))
                                .distinct()
                                .subtract(hits_with_subsession_id
                                          .select(col(self.hit_key_column), col(self.session_id_column)))
                                .join(hits_with_subsession_id
                                      .select(col(self.session_id_column), col(self.result_name))
                                      .distinct(), on=self.session_id_column, how='leftouter')
                                .select(col(self.session_id_column), col(self.hit_key_column), col(self.result_name))
                                )

        all_hits_with_subsession_id = (hits_with_subsession_id
                                       .union(not_timestamped_hits)
                                       .fillna({self.result_name: 1}))

        hits_with_subsession_id = (dataframe
                                   .where(col(self.report_group_type_column) == 'TIMESTAMP')
                                   .select(self.session_id_column, self.hit_key_column)
                                   .distinct()
                                   .join(hits_with_subsession_id, on=[self.session_id_column, self.hit_key_column], how='leftouter')
                                   .fillna({self.result_name: 1}))

        dataframe = (dataframe
                     .join(all_hits_with_subsession_id, on=[self.session_id_column, self.hit_key_column], how='leftouter'))

        return dataframe


class LastFactValueFinder(Transformer):
    """
    Gets the fact value of the last occurrence of a given event for each session.

    Note:
        The event must be timestamped through a report group of the timestamp type.

    Args:
        session_id_column: The column containing the session ids.
        event_id_column: The column containing the event ids.
        hit_key_column: The column containing the hit keys.
        fact_value_column: The column containing the fact values.
        timestamp_column: The column containing the timestamps.
        param_list: A list of named tuples containing the necessary params for each last fact value search. Default
            value is an empty list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns a named tuple type with all the params necessary to get the last fact values of an event. Ignore this
        if you are going to use the helper methods.
        """
        fields = [
            'event_id',
            'result_column_name'
        ]
        return namedtuple(cls.__name__ + 'TransformationParam', fields)

    def __init__(self, session_id_column, event_id_column, hit_key_column, fact_value_column, timestamp_column, param_list=None):
        if param_list is None:
            param_list = []
        super().__init__()
        self.session_id_column = session_id_column
        self.event_id_column = event_id_column
        self.hit_key_column = hit_key_column
        self.fact_value_column = fact_value_column
        self.timestamp_column = timestamp_column
        self.param_list = param_list

    def add_event(self, event_id, result_name):
        """
        Add a event to be searched for its last fact values.

        Args:
            event_id (str): The id of the event.
            result_name: The name of the result column containing the last fact values.

        Returns:
            This LastFactValueFinder, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(event_id=event_id, result_column_name=result_name)
        self.param_list.append(param)

    def transform(self, dataframe):
        """
        Performs the search of the last fact values.

        Args:
            dataframe: A DataFrame containing the timestamped events and its fact values.

        Returns:
             A new DataFrame containing the session id, as well as the obtained last fact values.
        """
        event_id_column = self.event_id_column
        session_id_column = self.session_id_column
        timestamp_column = self.timestamp_column
        hit_key_column = self.hit_key_column
        fact_value_column = self.fact_value_column
        sessions_df = dataframe.select(*session_id_column).distinct()

        events_to_tranform = {params.event_id: params.result_column_name
                              for params in self.param_list}

        # Genera la condición para filtrar las filas de los eventos que nos interesan
        hay_que_transformar_este_evento = reduce(
            lambda condicion1, condicion2: condicion1 + " OR " + condicion2,
            ['%s = "%s"' % (event_id_column, event_id) for event_id in events_to_tranform]
        )

        events_df = (dataframe
                     .where(hay_que_transformar_este_evento)
                     .where(col("ROW_TYPE") == "EVENT")
                     .select(*session_id_column, event_id_column, hit_key_column, fact_value_column))

        timestamps_df = (dataframe
                         .where(hay_que_transformar_este_evento)
                         .where('GROUP_TYPE = "TIMESTAMP"')
                         .select(*session_id_column, event_id_column, hit_key_column, timestamp_column)
                         .withColumnRenamed(timestamp_column, 'TIMESTAMP'))

        newest_timestamps = (timestamps_df
                             .groupBy(*session_id_column, event_id_column)
                             .max('TIMESTAMP')
                             .withColumnRenamed('max(TIMESTAMP)', 'TIMESTAMP'))

        newest_ocurrency = newest_timestamps.join(timestamps_df, on=['TIMESTAMP', event_id_column, *session_id_column], how='leftouter')

        last_fact_values = (newest_ocurrency
                            .join(events_df, on=[*session_id_column, event_id_column, 'HIT_KEY'], how="leftouter")
                            .select(*session_id_column, event_id_column, fact_value_column)
                            .groupBy(*session_id_column)
                            .pivot(event_id_column)
                            .agg(first(fact_value_column)))

        result = sessions_df.join(last_fact_values, on=session_id_column, how="leftouter")

        for event_id, column_name in events_to_tranform.items():
            if event_id in result.columns:
                result = result.withColumnRenamed(event_id, column_name)
            else:
                result = result.withColumn(column_name, lit(None).cast(StringType()))

        return result

    @property
    def session_id_column(self):
        return self.__session_id_column

    @session_id_column.setter
    def session_id_column(self, session_id_column):
        if type(session_id_column) is not list:
            self.__session_id_column = [session_id_column]
        else:
            self.__session_id_column = session_id_column


class FactValueCounter(Transformer):
    """
    For each id, counts the number of appearances of a given fact value in a given event.

    Args:
        session_id_column: The column containing the session ids.
        event_id_column: The column containing the event ids.
        row_type_column: The column containing the row type.
        param_list: A list of named tuples containing the necessary params for each fact value count. Default value is
            an empty list.
    """
    @classmethod
    def get_param_type(cls):
        """
         Returns a named tuple type with all the params necessary to perform a fact value count. Ignore this if you are
         going to use the helper methods.
         """
        fields = [
            'event_id',
            'fact_value',
            'result_column_name'
        ]
        return namedtuple(cls.__name__ + 'TransformationParam', fields)

    def __init__(self, session_id_column, event_id_column, row_type_column, param_list=None):
        if param_list is None:
            param_list = []
        super().__init__()
        self.session_id_column = session_id_column
        self.event_id_column = event_id_column
        self.row_type_column = row_type_column
        self.param_list = param_list

    def add_value_to_count(self, event_id, fact_value, result_name):
        """
        Add a value to be counted by the transform method.

        Args:
            event_id (str): The id of the event containing the fact value to be counted.
            fact_value: The fact value to be counted.
            result_name: The name of the result column containing the count.

        Returns:
            This FactValueCounter, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(event_id=event_id, fact_value=fact_value, result_column_name=result_name)
        self.param_list.append(param)

    def transform(self, dataframe):
        """
        Performs the fact value counts.

        Args:
            dataframe: A DataFrame containing the fact values to be counted.

        Returns:
             A new DataFrame containing the session id, as well as the count of each of the fact values for each session
        """
        event_id_column = self.event_id_column
        session_id_column = self.session_id_column
        sesiones = dataframe.select(*session_id_column).distinct()
        eventos = dataframe.where(col(self.row_type_column) == "EVENT")

        event_id_and_fact_value = [(params.event_id, params.fact_value) for params in self.param_list]

        hay_que_contar_este_value_de_este_evento = reduce(
            lambda condicion1, condicion2: '(%s) OR (%s)' % (condicion1, condicion2),
            ['%s = "%s" AND %s = "%s"' % (event_id_column, id_evento, "FACT_VALUE", fact_value)
             for id_evento, fact_value in event_id_and_fact_value]
        )

        # Hace la cuenta de eventos para cada sesión
        event_counts_by_session = (eventos
                                   .where(hay_que_contar_este_value_de_este_evento)
                                   .withColumn("tmp_EventsAndFactValues", concat(event_id_column, col("FACT_VALUE")))
                                   .groupBy(*session_id_column)
                                   .pivot("tmp_EventsAndFactValues")
                                   .count())

        result = (sesiones
                  .join(event_counts_by_session, on=session_id_column, how="left_outer")
                  )

        # Da a las columnas los nombres de los eventos
        for params in self.param_list:
            current_col_name = str(params.event_id) + str(params.fact_value)
            result = (result
                      .withColumnRenamed(current_col_name, params.result_column_name)
                      .fillna({params.result_column_name: 0}))

        return result

    @property
    def session_id_column(self):
        return self.__session_id_column

    @session_id_column.setter
    def session_id_column(self, session_id_column):
        if type(session_id_column) is not list:
            self.__session_id_column = [session_id_column]
        else:
            self.__session_id_column = session_id_column


class FactValueChecker(Transformer):
    """
    Checks if a fact value appears in a given event type for each session.

    Args:
        session_id_column: The column containing the session ids.
        event_id_column: The column containing the event ids.
        row_type_column: The column containing the row type.
        param_list: A list of named tuples containing the necessary params for each fact value check. Default value is
            an empty list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns a named tuple type with all the params necessary to check if a fact value appears for a session. Ignore
        this if you are going to use the helper methods.
        """
        fields = [
            'event_id',
            'fact_value',
            'result_column_name'
        ]
        return namedtuple(cls.__name__ + 'TransformationParam', fields)

    def __init__(self, session_id_column, event_id_column, row_type_column, param_list=None):
        if param_list is None:
            param_list = []
        super().__init__()
        self.session_id_column = session_id_column
        self.event_id_column = event_id_column
        self.row_type_column = row_type_column
        self.param_list = param_list

    def add_value_to_check(self, event_id, fact_value, result_name):
        """
        Sets a fact value of an event to be checked for each session.

        Args:
            event_id: The id of the event which fact values are to be checked.
            fact_value: The fact value to check.
            result_name: Te name of the result column.

        Returns:
            This FactValueChecker, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(event_id=event_id, fact_value=fact_value, result_column_name=result_name)
        self.param_list.append(param)
        return self

    def transform(self, dataframe):
        """
        Performs the fact value checks.

        Args:
            dataframe: A DataFrame containing the fact values to be checked.

        Returns:
             A new DataFrame containing the session id, as well as the check of each of the fact values for each session
        """
        event_id_column = self.event_id_column
        session_id_column = self.session_id_column
        sesiones = dataframe.select(*session_id_column).distinct()
        eventos = dataframe.where(col(self.row_type_column) == "EVENT")

        event_id_and_fact_value = [(params.event_id, params.fact_value) for params in self.param_list]

        hay_que_contar_este_value_de_este_evento = reduce(
            lambda condicion1, condicion2: '(%s) OR (%s)' % (condicion1, condicion2),
            ['%s = "%s" AND %s = "%s"' % (event_id_column, id_evento, "FACT_VALUE", fact_value)
             for id_evento, fact_value in event_id_and_fact_value]
        )

        # Hace la cuenta de eventos para cada sesión
        event_counts_by_session = (eventos
                                   .where(hay_que_contar_este_value_de_este_evento)
                                   .withColumn("tmp_EventsAndFactValues", concat(event_id_column, col("FACT_VALUE")))
                                   .groupBy(*session_id_column)
                                   .pivot("tmp_EventsAndFactValues")
                                   .agg(col("tmp_EventsAndFactValues").isNotNull()))

        result = (sesiones
                  .join(event_counts_by_session, on=session_id_column, how="left_outer")
                  )

        # Da a las columnas los nombres especificados
        for params in self.param_list:
            current_col_name = str(params.event_id) + str(params.fact_value)
            result = result.withColumnRenamed(current_col_name, params.result_column_name)

        booleanizer = udf(lambda x: 0 if x is None else 1)

        for params in self.param_list:
            if params.result_column_name in result.columns:
                result = (result
                          .withColumnRenamed(params.result_column_name, 'tmp_'+params.result_column_name)
                          .withColumn(params.result_column_name, booleanizer('tmp_'+params.result_column_name))
                          .drop('tmp_'+params.result_column_name)
                          .fillna({params.result_column_name: 0}))
            else:
                result = result.withColumn(params.result_column_name, lit(0))

        return result

    @property
    def session_id_column(self):
        return self.__session_id_column

    @session_id_column.setter
    def session_id_column(self, session_id_column):
        if type(session_id_column) is not list:
            self.__session_id_column = [session_id_column]
        else:
            self.__session_id_column = session_id_column

    
class FactValueFinder(Transformer):
    """
    Gets the fact value of a given event, for each session.

    Note:
        If that event occurs more that once for a session, any of the fact values will be taken, and just one of them.

    Args:
        session_id_column: The column containing the session ids.
        event_id_column: The column containing the event ids.
        param_list: A list of named tuples containing the necessary params for getting each fact value. Default value is
            an empty list.
    """
    @classmethod
    def get_param_type(cls):
        """
        Returns a named tuple type with all the params necessary to get fact values from an event. Ignore this if you
        are going to use the helper methods.
        """
        fields = [
            'event_id',
            'result_column_name'
        ]
        return namedtuple(cls.__name__ + 'TransformationParam', fields)

    def __init__(self, session_id_column, event_id_column, param_list=None):
        if param_list is None:
            param_list = []
        super().__init__()
        self.session_id_column = session_id_column
        self.event_id_column = event_id_column
        self.param_list = param_list

    def add_event(self, event_id, result_name):
        """
        Adds a event, so that its fact value will be obtained by the transform method.

        Args:
            event_id: The id of the event.
            result_name: The name of the result column.

        Returns:
            This FactValueFinder, so that method chaining can be used.
        """
        param_type = self.get_param_type()
        param = param_type(event_id=event_id, result_column_name=result_name)
        self.param_list.append(param)

    def transform(self, dataframe):
        """
        Gets the fact values for each of the given events.

        Args:
            dataframe: A DataFrame containing the fact values to be obtained.

        Returns:
            A new DataFrame containing the session id, as well as the the fact values for each of the given events.
        """
        event_id_column = self.event_id_column
        session_id_column = self.session_id_column
        events_to_tranform = {params.event_id: params.result_column_name
                              for params in self.param_list}

        # Genera la condición para filtrar las filas de los eventos que nos interesan
        hay_que_transformar_este_evento = reduce(
            lambda condicion1, condicion2: condicion1 + " OR " + condicion2,
            ['%s = "%s"' % (event_id_column, event_id) for event_id in events_to_tranform]
        )

        sesiones = dataframe.select(*session_id_column).distinct()
        eventos = (dataframe
                   .where(col("ROW_TYPE") == "EVENT")
                   .where(hay_que_transformar_este_evento)
                   .select(*session_id_column, event_id_column, 'FACT_VALUE'))

        fact_values = (eventos
                       .groupBy(*session_id_column)
                       .pivot(event_id_column)
                       .agg(first('FACT_VALUE')))

        result = sesiones.join(fact_values, on=session_id_column, how="leftouter")

        for event_id, column_name in events_to_tranform.items():
            if str(event_id) in result.columns:
                result = result.withColumnRenamed(event_id, column_name)
            else:
                result = result.withColumn(column_name, lit(None).cast(StringType()))

        return result

    @property
    def session_id_column(self):
        return self.__session_id_column

    @session_id_column.setter
    def session_id_column(self, session_id_column):
        if type(session_id_column) is not list:
            self.__session_id_column = [session_id_column]
        else:
            self.__session_id_column = session_id_column